@@ -1,10 +1,10 @@
-import argparse
 from collections import defaultdict
 from fabric.api import env
 import logging
-from snapshotting import BackupWorker
+from snapshotting import BackupWorker, RestoreWorker
 from snapshotting import Snapshot
 from snapshotting import SnapshotCollection
+from utils import base_parser
 
 
 def run_backup(args):
@@ -31,7 +31,10 @@ def run_backup(args):
         aws_access_key_id=args.aws_access_key_id,
         aws_secret_access_key=args.aws_secret_access_key,
         cassandra_data_path=args.cassandra_data_path,
-        nodetool_path=args.nodetool_path
+        nodetool_path=args.nodetool_path,
+        cassandra_bin_dir=args.cassandra_bin_dir,
+        backup_schema=args.backup_schema,
+        connection_pool_size=args.connection_pool_size
     )
 
     if create_snapshot:
@@ -68,33 +71,37 @@ def list_backups(args):
             print '\t %r hosts:%r keyspaces:%r table:%r' % (snapshot, snapshot.hosts, snapshot.keyspaces, snapshot.table)
         print '------------------------' + '-' * len(path)
 
+
+def restore_backup(args):
+    snapshots = SnapshotCollection(
+        args.aws_access_key_id,
+        args.aws_secret_access_key,
+        args.s3_base_path,
+        args.s3_bucket_name
+    )
+
+    if args.snapshot_name == 'LATEST':
+        snapshot = snapshots.get_latest()
+    else:
+        snapshot = snapshots.get_snapshot_by_name(args.backup_name)
+
+    worker = RestoreWorker(aws_access_key_id=args.aws_access_key_id,
+                           aws_secret_access_key=args.aws_secret_access_key,
+                           snapshot=snapshot)
+
+    if args.hosts:
+        hosts = args.hosts.split(',')
+    else:
+        hosts = snapshot.hosts
+
+    target_hosts = args.target_hosts.split(',')
+
+    worker.restore(args.keyspace, args.table, hosts, target_hosts)
+
+
 def main():
-    parser = argparse.ArgumentParser(
-        formatter_class=argparse.RawDescriptionHelpFormatter,
-        description=__doc__)
 
-    # common arguments
-    parser.add_argument('--aws-access-key-id',
-                        required=True,
-                        help='public AWS access key.')
-
-    parser.add_argument('--aws-secret-access-key',
-                        required=True,
-                        help='S3 secret access key.')
-
-    parser.add_argument('--s3-bucket-name',
-                        required=True,
-                        help='S3 bucket name for backups.')
-
-    parser.add_argument('--s3-base-path',
-                        required=True,
-                        help='S3 base path for backups.')
-
-    parser.add_argument('-v', '--verbose',
-                        action='store_true',
-                        help='increase output verbosity')
-
-    subparsers = parser.add_subparsers(title='subcommands',
+    subparsers = base_parser.add_subparsers(title='subcommands',
                                        dest='subcommand')
 
     subparsers.add_parser('list', help='list existing backups')
@@ -119,9 +126,13 @@ def main():
                                help='cassandra data path.')
 
     backup_parser.add_argument('--nodetool-path',
-                               default='/usr/bin/nodetool',
+                               default=None,
                                help='nodetool path.')
 
+    backup_parser.add_argument('--cassandra-bin-dir',
+                               default='/usr/bin',
+                               help='cassandra binaries directory')
+
     backup_parser.add_argument('--user',
                                help='the ssh user to loging on nodes')
 
@@ -129,7 +140,33 @@ def main():
                                action='store_true',
                                help='create a new snapshot')
 
-    args = parser.parse_args()
+    backup_parser.add_argument('--backup-schema',
+                               action='store_true',
+                               help='Backup (thrift) schema of selected keyspaces')
+
+    backup_parser.add_argument('--connection-pool-size',
+                               default=12,
+                               help='Number of simultaneous connections to cassandra nodes.')
+
+    # restore snapshot arguments
+    restore_parser = subparsers.add_parser('restore', help='restores a snapshot')
+    restore_parser.add_argument('--snapshot-name',
+                                default='LATEST',
+                                help='The name (date/time) of the snapshot (and incrementals) to restore')
+    restore_parser.add_argument('--keyspace',
+                                required=True,
+                                help='The keyspace to restore')
+    restore_parser.add_argument('--table',
+                                default='',
+                                help='The table (column family) to restore; leave blank for all')
+    restore_parser.add_argument('--hosts',
+                                default='',
+                                help='Comma separated list of hosts to restore from; leave empty for all')
+    restore_parser.add_argument('--target-hosts',
+                                required=True,
+                                help="The comma separated list of hosts to restore into")
+
+    args = base_parser.parse_args()
     subcommand = args.subcommand
 
     if args.verbose:
@@ -139,6 +176,8 @@ def main():
         run_backup(args)
     elif subcommand == 'list':
         list_backups(args)
+    elif subcommand == 'restore':
+        restore_backup(args)
 
 if __name__ == '__main__':
     main()
