@@ -1,3 +1,5 @@
+import re
+import shutil
 from boto.s3.connection import S3Connection
 from boto.s3.key import Key
 from datetime import datetime
@@ -7,11 +9,13 @@ from fabric.api import hide
 from fabric.api import put
 from fabric.api import sudo
 from fabric.context_managers import settings
+from multiprocessing.dummy import Pool
 import json
 import logging
 import os
 from tempfile import NamedTemporaryFile
 import time
+import sys
 
 
 class Snapshot(object):
@@ -88,15 +92,122 @@ class Snapshot(object):
     __str__ = __repr__
 
 
+class RestoreWorker(object):
+    def __init__(self, aws_access_key_id, aws_secret_access_key, snapshot):
+        self.aws_secret_access_key = aws_secret_access_key
+        self.aws_access_key_id = aws_access_key_id
+        self.s3connection = S3Connection(aws_access_key_id=self.aws_access_key_id,
+                                         aws_secret_access_key=self.aws_secret_access_key)
+        self.snapshot = snapshot
+        self.keyspace_table_matcher = None
+
+    def restore(self, keyspace, table, hosts, target_hosts):
+        # TODO:
+        # 4. sstableloader
+
+        logging.info("Restoring keyspace=%(keyspace)s, table=%(table)s" % dict(keyspace=keyspace,
+                                                                               table=table))
+
+        logging.info("From hosts: %(hosts)s to: %(target_hosts)s" % dict(hosts=', '.join(hosts),
+                                                                         target_hosts=', '.join(
+                                                                             target_hosts)))
+        if not table:
+            table = ".*?"
+
+        bucket = self.s3connection.get_bucket(self.snapshot.s3_bucket)
+
+        matcher_string = "(%(hosts)s).*/(%(keyspace)s)/(%(table)s)/" % dict(hosts='|'.join(hosts), keyspace=keyspace, table=table)
+        self.keyspace_table_matcher = re.compile(matcher_string)
+
+        keys = []
+        tables = set()
+
+        for k in bucket.list(self.snapshot.base_path):
+            r = self.keyspace_table_matcher.search(k.name)
+            if not r:
+                continue
+
+            tables.add(r.group(3))
+            keys.append(k)
+
+        self._delete_old_dir_and_create_new(keyspace, tables)
+
+        total_size = reduce(lambda s, k: s + k.size, keys, 0)
+
+        logging.info("Found %(files_count)d files, with total size of %(size)s." % dict(
+            files_count=len(keys),
+            size=self._human_size(total_size)))
+
+        self._download_keys(keys, total_size)
+
+        logging.info("Finished downloading...")
+
+        self._run_sstableloader(keyspace, tables, target_hosts)
+
+    def _delete_old_dir_and_create_new(self, keyspace, tables):
+        keyspace_path = './%s' % keyspace
+        if os.path.exists(keyspace_path) and os.path.isdir(keyspace_path):
+            logging.warning("Deleteing directory (%s)..." % keyspace_path)
+            shutil.rmtree(keyspace_path)
+
+        for table in tables:
+            path = './%s/%s' % (keyspace, table)
+            if not os.path.exists(path):
+                os.makedirs(path)
+
+    def _download_keys(self, keys, total_size, pool_size=5):
+        logging.info("Starting to download...")
+
+        progress_string = ""
+        read_bytes = 0
+
+        thread_pool = Pool(pool_size)
+
+        for size in thread_pool.imap(self._download_key, keys):
+            old_width = len(progress_string)
+            read_bytes += size
+            progress_string = "%s / %s (%.2f%%)" % (self._human_size(read_bytes),
+                                                    self._human_size(total_size),
+                                                    (read_bytes/float(total_size))*100.0)
+            width = len(progress_string)
+            padding = ""
+            if width < old_width:
+                padding = " "*(width-old_width)
+            progress_string = "%s%s\r" % (progress_string, padding)
+
+            sys.stderr.write(progress_string)
+
+    def _download_key(self, key):
+        r = self.keyspace_table_matcher.search(key.name)
+        filename = './%s/%s/%s_%s' % (r.group(2), r.group(3), key.name.split('/')[2], key.name.split('/')[-1])
+        key.get_contents_to_filename(filename)
+
+        return key.size
+
+    def _human_size(self, size):
+        for x in ['bytes', 'KB', 'MB', 'GB']:
+            if size < 1024.0:
+                return "%3.1f%s" % (size, x)
+            size /= 1024.0
+        return "%3.1f%s" % (size, 'TB')
+
+    def _run_sstableloader(self, keyspace, tables, target_hosts):
+        # TODO: get path to sstableloader
+        for table in tables:
+            command = 'sstableloader --nodes %(hosts)s -v %(keyspace)s/%(table)s' % dict(
+                hosts=','.join(target_hosts), keyspace=keyspace, table=table)
+            logging.info("invoking: %s", command)
+
+            os.system(command)
+
+
 class BackupWorker(object):
     """
-    BackupWorker does the actual snapshot / backup work
-
     Backup process is split in this steps:
-        - requesting cassandra to create new backups
-        - uploadint backup files to S3
-        - clearing backup files from nodes
-        - updating backup meta informations
+        - requests cassandra to create new backups
+        - uploads backup files to S3
+        - clears backup files from nodes
+        - updates backup meta informations
 
     When performing a new snapshot the manifest of the snapshot is
     uploaded to S3 for later use.
@@ -109,14 +220,17 @@ class BackupWorker(object):
 
     """
 
-    connection_pool_size = 12
-
     def __init__(self, aws_secret_access_key,
-                 aws_access_key_id, cassandra_data_path, nodetool_path):
+                 aws_access_key_id, cassandra_data_path,
+                 nodetool_path, cassandra_bin_dir, backup_schema,
+                 connection_pool_size=12):
         self.aws_secret_access_key = aws_secret_access_key
         self.aws_access_key_id = aws_access_key_id
         self.cassandra_data_path = cassandra_data_path
-        self.nodetool_path = nodetool_path
+        self.nodetool_path = nodetool_path or "%s/nodetool" % cassandra_bin_dir
+        self.cassandra_cli_path = "%s/cassandra-cli" % cassandra_bin_dir
+        self.backup_schema = backup_schema
+        self.connection_pool_size = connection_pool_size
 
     def get_current_node_hostname(self):
         return env.host_string
@@ -124,7 +238,7 @@ class BackupWorker(object):
     def get_remote_tmp_folder(self):
         return '/tmp/'
 
-    def create_s3funnel_manifest(self, files):
+    def create_upload_manifest(self, files):
         manifest = NamedTemporaryFile(delete=False)
         manifest.write('\n'.join("%s" % f for f in files))
         manifest.close()
@@ -133,12 +247,12 @@ class BackupWorker(object):
     def upload_backups_to_s3(self, snapshot, files):
         prefix = '/'.join(snapshot.base_path.split(
             '/') + [self.get_current_node_hostname()])
-        manifest = self.create_s3funnel_manifest(files)
+        manifest = self.create_upload_manifest(files)
         manifest_path = self.get_remote_tmp_folder() + manifest.name.split(os.path.sep)[-1]
         put(manifest.name, manifest_path)
         os.unlink(manifest.name)
 
-        upload_command = "s3funnel %(bucket)s PUT --put-full-path --threads 4 --add-prefix %(prefix)s --aws_key=%(key)s --aws_secret=%(secret)s --input=%(manifest)s -v"
+        upload_command = "cassandra-snapshotter-agent --aws-access-key-id=%(key)s --aws-secret-access-key=%(secret)s --s3-bucket-name=%(bucket)s --s3-base-path=%(prefix)s  put --manifest=%(manifest)s --concurrency=4"
         cmd = upload_command % dict(
             bucket=snapshot.s3_bucket,
             prefix=prefix,
@@ -161,6 +275,8 @@ class BackupWorker(object):
             self.clear_cluster_snapshot(snapshot)
         self.write_ring_description(snapshot)
         self.write_snapshot_manifest(snapshot)
+        if self.backup_schema:
+            self.write_schema(snapshot)
 
     def update_snapshot(self, snapshot):
         """
@@ -173,13 +289,27 @@ class BackupWorker(object):
         finally:
             self.clear_cluster_backups(snapshot)
         self.write_ring_description(snapshot)
+        if self.backup_schema:
+            self.write_schema(snapshot)
 
     def get_ring_description(self):
         with settings(host_string=env.hosts[0]):
             with hide('output'):
-                ring_description = sudo('nodetool ring')
+                ring_description = sudo(self.nodetool_path + ' ring')
         return ring_description
 
+    def get_keyspace_schema(self, keyspace=None):
+        output = ""
+        with settings(host_string=env.hosts[0]):
+            with hide('output'):
+                cmd = "echo -e 'show schema;\n' | %s" % (self.cassandra_cli_path)
+                if keyspace:
+                    cmd = "echo -e 'show schema;\n' | %s -k %s" % (self.cassandra_cli_path, keyspace)
+                output = sudo(cmd)
+        # remove unwanted lines
+        schema = '\n'.join([l for l in output.split("\n") if re.match(r'(create|use| )',l)])
+        return schema
+
     def write_on_S3(self, bucket_name, path, content):
         conn = S3Connection(self.aws_access_key_id, self.aws_secret_access_key)
         bucket = conn.get_bucket(bucket_name)
@@ -187,10 +317,24 @@ class BackupWorker(object):
         key.set_contents_from_string(content)
 
     def write_ring_description(self, snapshot):
+        logging.info('Writing ring description')
         content = self.get_ring_description()
         ring_path = '/'.join([snapshot.base_path, 'ring'])
         self.write_on_S3(snapshot.s3_bucket, ring_path, content)
 
+    def write_schema(self, snapshot):
+        if snapshot.keyspaces:
+            for ks in snapshot.keyspaces.split(","):
+                logging.info('Writing schema for keyspace %s' % ks)
+                content = self.get_keyspace_schema(ks)
+                schema_path = '/'.join([snapshot.base_path, "schema_%s.cdl" % ks])
+                self.write_on_S3(snapshot.s3_bucket, schema_path, content)
+        else:
+            logging.info('Writing schema for all keyspaces')
+            content = self.get_keyspace_schema()
+            schema_path = '/'.join([snapshot.base_path, "schema.cdl"])
+            self.write_on_S3(snapshot.s3_bucket, schema_path, content)
+
     def write_snapshot_manifest(self, snapshot):
         content = snapshot.dump_manifest_file()
         manifest_path = '/'.join([snapshot.base_path, 'manifest.json'])
@@ -214,7 +358,7 @@ class BackupWorker(object):
         if incremental_backups:
             backup_command = '%(nodetool)s flush %(keyspaces)s %(table_param)s'
         else:
-            backup_command = '%(nodetool)s snapshot -t "%(snapshot)s" %(keyspaces)s %(table_param)s'
+            backup_command = '%(nodetool)s snapshot -t %(snapshot)s %(keyspaces)s %(table_param)s'
 
         cmd = backup_command % dict(
             nodetool=self.nodetool_path,
@@ -314,6 +458,9 @@ class SnapshotCollection(object):
         self.aws_secret_access_key = aws_secret_access_key
 
     def _read_s3(self):
+        if self.snapshots:
+            return
+
         conn = S3Connection(self.aws_access_key_id, self.aws_secret_access_key)
         bucket = conn.get_bucket(self.s3_bucket)
         self.snapshots = []
@@ -331,6 +478,14 @@ class SnapshotCollection(object):
                 Snapshot.load_manifest_file(manifest_data, self.s3_bucket))
         self.snapshots = sorted(self.snapshots, reverse=True)
 
+    def get_snapshot_by_name(self, name):
+        snapshots = filter(lambda s: s.name == name, self)
+        return snapshots and snapshots[0]
+
+    def get_latest(self):
+        self._read_s3()
+        return self.snapshots[0]
+
     def get_snapshot_for(self, hosts, keyspaces, table):
         '''
         returns the most recent compatible snapshot
@@ -345,6 +500,5 @@ class SnapshotCollection(object):
             return snapshot
 
     def __iter__(self):
-        if self.snapshots is None:
-            self._read_s3()
+        self._read_s3()
         return iter(self.snapshots)
